from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import attr

from serena.payloads.header import BasicHeader
from serena.payloads.method import (
    BasicDeliverPayload,
    BasicGetOkPayload,
    MethodPayload,
    method_payload_name,
)

if TYPE_CHECKING:
    from serena.channel import Channel


@attr.s(frozen=True, slots=True)
class AMQPEnvelope:
    """
    Wraps metadata related to the *delivery* of the message.
    """

    #: The identifier for the consumer. May be None if this message was gotten synchronously.
    consumer_tag: Optional[str] = attr.ib()

    #: The server-assigned delivery tag.
    delivery_tag: int = attr.ib()

    #: Indicates that the message has been previously delivered.
    redelivered: bool = attr.ib()

    #: The name of the exchange the message was originally published to.
    exchange_name: str = attr.ib()

    #: The routing key for the message.
    routing_key: str = attr.ib()

    #: The messages remaining in the queue. May be None if this message was consumed asynchronously.
    message_count: Optional[int] = attr.ib()

    @classmethod
    def from_deliver(cls, payload: BasicDeliverPayload) -> AMQPEnvelope:
        """
        Creates a new envelope from a :class:`.BasicDeliverPayload.`
        """

        return AMQPEnvelope(
            consumer_tag=payload.consumer_tag,
            delivery_tag=payload.delivery_tag,
            redelivered=payload.redelivered,
            exchange_name=payload.exchange_name,
            routing_key=payload.routing_key,
            message_count=None,
        )

    @classmethod
    def from_get(cls, payload: BasicGetOkPayload) -> AMQPEnvelope:
        """
        Creates a new envelope from a :class:`.BasicGetOkPayload`.
        """

        return AMQPEnvelope(
            consumer_tag=None,
            delivery_tag=payload.delivery_tag,
            redelivered=payload.redelivered,
            exchange_name=payload.exchange_name,
            routing_key=payload.routing_key,
            message_count=payload.message_count,
        )

    @classmethod
    def of(cls, payload: MethodPayload) -> AMQPEnvelope:
        """
        Creates a new envelope from a method payload.
        """

        if isinstance(payload, BasicDeliverPayload):
            return cls.from_deliver(payload)

        elif isinstance(payload, BasicGetOkPayload):
            return cls.from_get(payload)

        else:
            raise TypeError(
                f"Expected Basic.Deliver or Basic.Get-Ok, got {method_payload_name(payload)}"
            )


@attr.s(frozen=True, slots=True)
class AMQPMessage:
    """
    The wrapper around a single, delivered AMQP message.
    """

    _channel: Channel = attr.ib()

    #: The "envelope" for the message. Wraps data about the delivery of the message.
    envelope: AMQPEnvelope = attr.ib()

    #: The header for the message, containing application-specific details.
    header: BasicHeader = attr.ib()

    #: The actual body of this message.
    body: bytes = attr.ib()

    async def ack(self, *, multiple: bool = False):
        """
        Acknowledges this message. See :class:`~.Channel.basic_ack`.
        """

        return await self._channel.basic_ack(self.envelope.delivery_tag, multiple=multiple)

    async def nack(self, *, multiple: bool = False, requeue: bool = True):
        """
        Negatively acknowledges this message. See :class:`~.Channel.basic_nack`.
        """

        return await self._channel.basic_nack(
            self.envelope.delivery_tag, multiple=multiple, requeue=requeue
        )

    async def reject(self, *, requeue: bool = True):
        """
        Rejects this message. See :class:`~.Channel.basic_reject`.
        """

        return await self._channel.basic_reject(self.envelope.delivery_tag, requeue=requeue)



import logging

# our public exports, relatively minimal
from serena.channel import Channel
from serena.connection import AMQPConnection, open_connection
from serena.enums import *
from serena.exc import *
from serena.message import AMQPMessage
from serena.mixin import ChannelDelegate, ChannelLike
from serena.payloads.header import BasicHeader

if not hasattr(logging.Logger, "trace"):
    # TRACE_LEVEL = 5
    logging.addLevelName(5, "TRACE")

    def trace(self, message, *args, **kws):
        if self.isEnabledFor(5):
            self._log(5, message, args, **kws)

    logging.Logger.trace = trace
    del trace

del logging



import abc
from typing import Any, AsyncContextManager, AsyncIterable, Optional, Union

from serena.enums import ExchangeType
from serena.message import AMQPMessage
from serena.payloads.header import BasicHeader
from serena.payloads.method import QueueDeclareOkPayload


class ChannelLike(abc.ABC):
    """
    Base object shared between the :class:`.Channel` and :class:`.ChannelPool` object.
    """

    @abc.abstractmethod
    async def exchange_declare(
        self,
        name: str,
        type: Union[ExchangeType, str],
        *,
        passive: bool = False,
        durable: bool = False,
        auto_delete: bool = False,
        internal: bool = False,
        **arguments: Any,
    ) -> str:
        """
        Declares a new exchange.

        :param name: The name of the exchange. Must not be empty.
        :param type: The type of the exchange to create.
        :param passive: If True, the server will return a DeclareOk if the exchange exists, and
                        an error if it doesn't. This can be used to inspect server state without
                        modification.
        :param durable: If True, then the declared exchange will survive a server restart.
        :param auto_delete: If True, then the declared exchange will be automatically deleted
                            when all queues have finished using it.
        :param internal: If True, then the exchange may not be used directly by publishers.
        :param arguments: A dictionary of implementation-specific arguments.
        :return: The name of the exchange, as it exists on the server.
        """

    @abc.abstractmethod
    async def exchange_delete(
        self,
        name: str,
        *,
        if_unused: bool = False,
    ) -> None:
        """
        Deletes an exchange.

        :param name: The name of the exchange to delete.
        :param if_unused: If True, then the exchange will only be deleted if it has no queue
                          bindings.
        :return: Nothing.
        """

    @abc.abstractmethod
    async def exchange_bind(
        self,
        destination: str,
        source: str,
        routing_key: str,
        **arguments: Any,
    ) -> None:
        """
        Binds an exchange to another exchange. This is a
        `RabbitMQ extension <https://www.rabbitmq.com/e2e.html>`__ and may not be supported in other
        AMQP implementations.

        :param destination: The name of the destination exchange to bind. A blank name means the
                            default exchange.
        :param source: The name of the source exchange to bind. A blank name means the default
                       exchange.
        :param routing_key: The routing key for the exchange binding.
        :param arguments: A dictionary of implementation-specific arguments.
        :return: Nothing.
        """

    @abc.abstractmethod
    async def exchange_unbind(
        self,
        destination: str,
        source: str,
        routing_key: str,
        **arguments: Any,
    ) -> None:
        """
        Unbinds an exchange from another exchange. This is a
        `RabbitMQ extension <https://www.rabbitmq.com/e2e.html>`__ and may not be supported in other
        AMQP implementations.

        :param destination: The name of the destination exchange to unbind. A blank name means the
                            default exchange.
        :param source: The name of the source exchange to unbind. A blank name means the default
                       exchange.
        :param routing_key: The routing key for the exchange binding that is being unbinded.
        :param arguments: A dictionary of implementation-specific arguments.
        :return: Nothing.
        """

    @abc.abstractmethod
    async def queue_declare(
        self,
        name: str,
        *,
        passive: bool = False,
        durable: bool = False,
        exclusive: bool = False,
        auto_delete: bool = False,
        **arguments: Any,
    ) -> QueueDeclareOkPayload:
        """
        Declares a queue.

        :param name: The name of the queue. If blank, a name will be automatically generated by
                     the server and returned.
        :param passive: If True, the server will return a DeclareOk if the queue exists, and an
                        error if it doesn't. This can be used to inspect server state without
                        modification.
        :param durable: If True, the queue being created will persist past server restarts.
        :param exclusive: If True, this queue will only belong to this connection, and will be
                          automatically deleted when the connection closes. Best combined with an
                          automatically generated queue name.
        :param auto_delete: If True, this queue will be automatically deleted after all consumers
                            have finished. The queue will never be deleted before the first consumer
                            starts.
        :param arguments: Optional server implementation-specific arguments.
        :return: The :class:`.QueueDeclareOkPayload` the server returned.
        """

    @abc.abstractmethod
    async def queue_bind(
        self,
        queue_name: str,
        exchange_name: str,
        routing_key: str,
        **arguments: Any,
    ) -> None:
        """
        Binds a queue to an exchange.

        :param queue_name: The queue to bind.
        :param exchange_name: The exchange to bind to.
        :param routing_key: The routing key to use when binding.
        :param arguments: Any server-specific or exchange-specific extra arguments.
        :return: Nothing.
        """

    @abc.abstractmethod
    async def queue_delete(
        self,
        queue_name: str,
        *,
        if_empty: bool = False,
        if_unused: bool = False,
    ) -> int:
        """
        Deletes a queue.

        :param queue_name: The name of the queue to delete.
        :param if_empty: If True, the queue will only be deleted if it is empty.
        :param if_unused: If True, the queue will only be deleted if it is unused.
        :return: The number of messages deleted.
        """

    @abc.abstractmethod
    async def queue_purge(
        self,
        queue_name: str,
    ) -> int:
        """
        Purges all messages from a queue.

        :param queue_name: The name of the queue to be purged.
        :return: The number of messages deleted.
        """

    @abc.abstractmethod
    async def queue_unbind(
        self,
        queue_name: str,
        exchange_name: str,
        routing_key: str,
        **arguments: Any,
    ) -> None:
        """
        Unbinds a queue from an exchange.

        :param queue_name: The name of the queue to unbind.
        :param exchange_name: The name of the exchange to unbind from.
        :param routing_key: The routing key to unbind using.
        :param arguments: Implementation-specific arguments to use.
        """

    @abc.abstractmethod
    def basic_consume(
        self,
        queue_name: str,
        consumer_tag: str = "",
        *,
        no_local: bool = False,
        no_ack: bool = False,
        exclusive: bool = False,
        auto_ack: bool = True,
        **arguments: Any,
    ) -> AsyncContextManager[AsyncIterable[AMQPMessage]]:
        """
        Starts a basic consume operation. This returns an async context manager over an asynchronous
        iterator that yields incoming :class:`.AMQPMessage` instances.

        The channel can still be used for other operations during this operation.

        :param queue_name: The name of the queue to consume from.
        :param consumer_tag: The tag for this consume.
        :param no_local: If True, messages will not be sent to this consumer if it is on the same
                         connection that published them.
        :param no_ack: If True, messages will not be expected to be acknowledged. This can cause
                       data loss.
        :param exclusive: If True, then only this consumer can access the queue. Will fail if there
                          is another consumer already active.
        :param arguments: Implementation-specific arguments.
        :param auto_ack: If True, then messages will be automatically positively acknowledged
                         in the generator loop. Has no effect if ``no_ack`` is True. This is a
                         Serena-exclusive feature, not a protocol feature.
        """

    @abc.abstractmethod
    async def basic_publish(
        self,
        exchange_name: str,
        routing_key: str,
        body: bytes,
        *,
        header: BasicHeader = None,
        mandatory: bool = True,
        immediate: bool = False,
    ):
        """
        Publishes a message to a specific exchange.

        :param exchange_name: The name of the exchange to publish to. This can be blank to mean the
                              default exchange.
        :param routing_key: The routing key to publish to.
        :param body: The body for this payload.
        :param header: The headers to use for this message. If unset, will use the default blank
                       headers.
        :param mandatory: Iff True, the server must return a ``Return`` message if the message
                          could not be routed to a queue.
        :param immediate: Iff True, the server must return a ``Return`` message if the message could
                          not be immediately consumed.
        :raise MessageReturnedError: If the message was returned to the publisher.

        .. warning::

            The immediate flag is *not* supported in RabbitMQ 3.x, and will cause the connection
            to close.
        """

    @abc.abstractmethod
    async def basic_get(self, queue: str, *, no_ack: bool = False) -> Optional[AMQPMessage]:
        """
        Gets a single message from a queue.

        :param queue: The queue to get the message from.
        :param no_ack: Iff not True, then messages will need to be explicitly acknowledged on
                       consumption.
        :return: A :class:`.AMQPMessage` if one existed on the queue, otherwise None.
        """


class ChannelDelegate(ChannelLike):  # pragma: no cover
    """
    Helper object that is a :class:`.ChannelLike` but is implemented via wrapping another channel.
    This is useful for being inherited from and adding new methods for convenience shortcuts.
    """

    def __init__(self, channel: ChannelLike):
        """
        :param channel: The :class:`.ChannelLike` to delegate to.
        """

        self._delegate = channel

    async def exchange_declare(
        self,
        name: str,
        type: Union[ExchangeType, str],
        *,
        passive: bool = False,
        durable: bool = False,
        auto_delete: bool = False,
        internal: bool = False,
        **arguments: Any,
    ) -> str:
        return await self._delegate.exchange_declare(
            name=name,
            type=type,
            passive=passive,
            durable=durable,
            auto_delete=auto_delete,
            internal=internal,
            **arguments,
        )

    async def exchange_delete(self, name: str, *, if_unused: bool = False) -> None:
        return await self._delegate.exchange_delete(name, if_unused=if_unused)

    async def exchange_bind(
        self, destination: str, source: str, routing_key: str, **arguments: Any
    ) -> None:
        return await self._delegate.exchange_bind(
            destination=destination, source=source, routing_key=routing_key, **arguments
        )

    async def exchange_unbind(
        self, destination: str, source: str, routing_key: str, **arguments: Any
    ) -> None:
        return await self._delegate.exchange_unbind(
            destination=destination, source=source, routing_key=routing_key, **arguments
        )

    async def queue_declare(
        self,
        name: str = "",
        *,
        passive: bool = False,
        durable: bool = False,
        exclusive: bool = False,
        auto_delete: bool = False,
        **arguments: Any,
    ) -> QueueDeclareOkPayload:
        return await self._delegate.queue_declare(
            name=name,
            passive=passive,
            durable=durable,
            exclusive=exclusive,
            auto_delete=auto_delete,
            **arguments,
        )

    async def queue_bind(
        self, queue_name: str, exchange_name: str, routing_key: str, **arguments: Any
    ) -> None:
        return await self._delegate.queue_bind(
            queue_name=queue_name, exchange_name=exchange_name, routing_key=routing_key, **arguments
        )

    async def queue_delete(
        self, queue_name: str, *, if_empty: bool = False, if_unused: bool = False
    ) -> int:
        return await self._delegate.queue_delete(
            queue_name=queue_name, if_empty=if_empty, if_unused=if_unused
        )

    async def queue_purge(self, queue_name: str) -> int:
        return await self._delegate.queue_purge(queue_name=queue_name)

    async def queue_unbind(
        self, queue_name: str, exchange_name: str, routing_key: str, **arguments: Any
    ) -> None:
        return await self._delegate.queue_unbind(
            queue_name=queue_name, exchange_name=exchange_name, routing_key=routing_key, **arguments
        )

    async def basic_ack(self, delivery_tag: int, *, multiple: bool = False):
        return await self._delegate.basic_ack(delivery_tag=delivery_tag, multiple=multiple)

    async def basic_reject(self, delivery_tag: int, *, requeue: bool = True):
        return await self._delegate.basic_reject(delivery_tag=delivery_tag, requeue=requeue)

    async def basic_nack(self, delivery_tag: int, *, multiple: bool = False, requeue: bool = False):
        return await self._delegate.basic_nack(
            delivery_tag=delivery_tag, multiple=multiple, requeue=requeue
        )

    def basic_consume(
        self,
        queue_name: str,
        consumer_tag: str = "",
        *,
        no_local: bool = False,
        no_ack: bool = False,
        exclusive: bool = False,
        auto_ack: bool = True,
        **arguments: Any,
    ) -> AsyncContextManager[AsyncIterable[AMQPMessage]]:
        return self._delegate.basic_consume(
            queue_name=queue_name,
            consumer_tag=consumer_tag,
            no_local=no_local,
            no_ack=no_ack,
            exclusive=exclusive,
            auto_ack=auto_ack,
            **arguments,
        )

    async def basic_publish(
        self,
        exchange_name: str,
        routing_key: str,
        body: bytes,
        *,
        header: BasicHeader = None,
        mandatory: bool = True,
        immediate: bool = False,
    ):
        return await self._delegate.basic_publish(
            exchange_name=exchange_name,
            routing_key=routing_key,
            body=body,
            header=header,
            mandatory=mandatory,
            immediate=immediate,
        )

    async def basic_get(self, queue: str, *, no_ack: bool = False) -> Optional[AMQPMessage]:
        return await self._delegate.basic_get(queue=queue, no_ack=no_ack)



from math import ceil


class BitSet(object):
    """
    An indexable, compact array of booleans that is stored using integers.
    """

    def __init__(self, size: int):
        self._items = [0] * ceil(size / 64)

    def _get_index(self, idx: int) -> int:
        index = idx // 64
        if index > len(self._items):
            raise IndexError(f"index {idx} out of range (max: {len(self._items) * 64})") from None

        return index

    def __len__(self):
        return len(self._items) * 64

    def __iter__(self):
        for idx in range(0, len(self)):
            yield self[idx]

    def __setitem__(self, key: int, value: bool):
        index = self._get_index(key)
        item = self._items[index]

        bit_idx = key % 64
        item &= ~(1 << bit_idx)
        if value:
            item |= 1 << bit_idx

        self._items[index] = item

    def __getitem__(self, item: int) -> bool:
        index = self._get_index(item)
        bit_idx = item % 64
        return (self._items[index] & (1 << bit_idx)) != 0



from __future__ import annotations

import abc
import enum
from typing import ClassVar

import attr


class FrameType(enum.IntEnum):
    """
    Enumeration of possible frame types.
    """

    METHOD = 1
    HEADER = 2
    BODY = 3
    HEARTBEAT = 4


@attr.s(frozen=True, slots=True)
class Frame(abc.ABC):
    """
    Base class for frame objects.
    """

    #: The type of this frame.
    type: ClassVar[FrameType]

    #: The channel ID of this frame.
    channel_id: int = attr.ib()


@attr.s(frozen=True, slots=True)
class HeartbeatFrame(Frame):
    """
    A single heartbeat frame. This should have an empty body.
    """

    type = FrameType.HEARTBEAT


@attr.s(frozen=True, slots=True)
class BodyFrame(Frame):
    """
    A single body frame.
    """

    type = FrameType.BODY

    data: bytes = attr.ib()



from __future__ import annotations

from contextlib import asynccontextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    AsyncIterable,
    Optional,
    Union,
)

import anyio
from anyio import Event, WouldBlock

from serena.channel import Channel
from serena.enums import ExchangeType
from serena.exc import UnexpectedCloseError
from serena.message import AMQPMessage
from serena.mixin import ChannelLike
from serena.payloads.header import BasicHeader
from serena.payloads.method import QueueDeclareOkPayload

if TYPE_CHECKING:
    from serena.connection import AMQPConnection


class ChannelPool(ChannelLike):
    """
    A pooled container of multiple channels that simplifies both usage and error handling.
    """

    def __init__(self, connection: AMQPConnection, initial_size: int = 64):
        """
        :param connection: The :class:`.AMQPConnection` to use.
        :param initial_size: The initial size of the pool top use. It may grow.
        """

        self._conn = connection
        self._qwrite, self._qread = anyio.create_memory_object_stream(
            max_buffer_size=initial_size, item_type=Channel
        )

        self._needs_new_connection = Event()

    @property
    def idle_channels(self) -> int:
        """
        Returns the number of idle channels.
        """

        return self._qread.statistics().current_buffer_used

    @property
    def open_channels(self) -> int:
        """
        Returns the number of open channels.
        """

        return int(self._qread.statistics().max_buffer_size)

    def _return_channel(self, channel):
        self._qwrite.send_nowait(channel)

    async def _open(self, initial_size: int):
        """
        Opens the pool and adds channels.
        """

        for _ in range(0, initial_size):
            # noinspection PyAsyncCall
            self._qwrite.send_nowait(await self._conn._open_channel())

    async def _close(self):
        """
        Closes all channels in the pool.
        """

        async with anyio.create_task_group() as nursery:
            while True:
                try:
                    next_channel = self._qread.receive_nowait()
                except WouldBlock:
                    self._qread.close()
                    break

                nursery.start_soon(next_channel.wait_until_closed)
                await self._conn._close_channel(next_channel.id)

    async def _open_channels(self):
        """
        Opens channels in an infinite loop.
        """

        while True:
            await self._needs_new_connection.wait()
            channel = await self._conn._open_channel()
            # noinspection PyAsyncCall
            self._qwrite.send_nowait(channel)

    # re-exported so that pycharm likes the type annotation better
    def checkout(self) -> AsyncContextManager[Channel]:
        """
        Checks out a new channel from the pool, and uses it persistently. The channel lifetime
        will be automatically managed for you.
        """

        return self._checkout()

    @asynccontextmanager
    async def _checkout(self):
        channel = await self._qread.receive()

        try:
            yield channel
        except UnexpectedCloseError:
            # noinspection PyAsyncCall
            self._needs_new_connection.set()
            self._needs_new_connection = Event()
            raise

        # gross!
        except BaseException:
            self._return_channel(channel)
            raise
        else:
            self._return_channel(channel)

    def basic_consume(
        self,
        queue_name: str,
        consumer_tag: str = "",
        *,
        no_local: bool = False,
        no_ack: bool = False,
        exclusive: bool = False,
        auto_ack: bool = True,
        **arguments: Any
    ) -> AsyncContextManager[AsyncIterable[AMQPMessage]]:
        """
        Starts a basic consume operation. This returns an async context manager over an asynchronous
        iterator that yields incoming :class:`.AMQPMessage` instances.
        The channel can still be used for other operations during this operation.

        :param queue_name: The name of the queue to consume from.
        :param consumer_tag: The tag for this consume.
        :param no_local: If True, messages will not be sent to this consumer if it is on the same
                         connection that published them.
        :param no_ack: If True, messages will not be expected to be acknowledged. This can cause
                       data loss.
        :param exclusive: If True, then only this consumer can access the queue. Will fail if there
                          is another consumer already active.
        :param arguments: Implementation-specific arguments.
        :param auto_ack: If True, then messages will be automatically positively acknowledged
                         in the generator loop. Has no effect if ``no_ack`` is True. This is a
                         Serena-exclusive feature, not a protocol feature.
        """

        @asynccontextmanager
        async def _do():
            async with self._checkout() as channel:
                async with channel.basic_consume(
                    queue_name=queue_name,
                    consumer_tag=consumer_tag,
                    no_local=no_local,
                    no_ack=no_ack,
                    exclusive=exclusive,
                    auto_ack=auto_ack,
                    **arguments,
                ) as it:
                    yield it

        return _do()

    async def basic_get(self, queue: str, *, no_ack: bool = False) -> Optional[AMQPMessage]:
        """
        Gets a single message from a queue.

        :param queue: The queue to get the message from.
        :param no_ack: Iff not True, then messages will need to be explicitly acknowledged on
                       consumption.
        :return: A :class:`.AMQPMessage` if one existed on the queue, otherwise None.
        """

        async with self._checkout() as channel:
            return await channel.basic_get(
                queue=queue,
                no_ack=no_ack,
            )

    async def basic_publish(
        self,
        exchange_name: str,
        routing_key: str,
        body: bytes,
        *,
        header: BasicHeader = None,
        mandatory: bool = True,
        immediate: bool = False
    ):
        """
        Publishes a message to a specific exchange.

        :param exchange_name: The name of the exchange to publish to. This can be blank to mean the
                              default exchange.
        :param routing_key: The routing key to publish to.
        :param body: The body for this payload.
        :param header: The headers to use for this message. If unset, will use the default blank
                       headers.
        :param mandatory: Iff True, the server must return a ``Return`` message if the message
                          could not be routed to a queue.
        :param immediate: Iff True, the server must return a ``Return`` message if the message could
                          not be immediately consumed.
        :raise MessageReturnedError: If the message was returned to the publisher.

        .. warning::
            The immediate flag is *not* supported in RabbitMQ 3.x, and will cause the connection
            to close.
        """

        async with self._checkout() as channel:
            await channel.basic_publish(
                exchange_name=exchange_name,
                routing_key=routing_key,
                body=body,
                header=header,
                mandatory=mandatory,
                immediate=immediate,
            )

    async def exchange_bind(
        self, destination: str, source: str, routing_key: str, **arguments: Any
    ) -> None:
        """
        Binds an exchange to another exchange. This is a
        `RabbitMQ extension <https://www.rabbitmq.com/e2e.html>`__ and may not be supported in other
        AMQP implementations.

        :param destination: The name of the destination exchange to bind. A blank name means the
                            default exchange.
        :param source: The name of the source exchange to bind. A blank name means the default
                       exchange.
        :param routing_key: The routing key for the exchange binding.
        :param arguments: A dictionary of implementation-specific arguments.
        :return: Nothing.
        """

        async with self._checkout() as channel:
            await channel.exchange_bind(
                destination=destination,
                source=source,
                routing_key=routing_key,
                arguments=arguments,
            )

    async def exchange_declare(
        self,
        name: str,
        type: Union[ExchangeType, str],
        *,
        passive: bool = False,
        durable: bool = False,
        auto_delete: bool = False,
        internal: bool = False,
        **arguments: Any
    ) -> str:
        """
        Declares a new exchange.

        :param name: The name of the exchange. Must not be empty.
        :param type: The type of the exchange to create.
        :param passive: If True, the server will return a DeclareOk if the exchange exists, and
                        an error if it doesn't. This can be used to inspect server state without
                        modification.
        :param durable: If True, then the declared exchange will survive a server restart.
        :param auto_delete: If True, then the declared exchange will be automatically deleted
                            when all queues have finished using it.
        :param internal: If True, then the exchange may not be used directly by publishers.
        :param arguments: A dictionary of implementation-specific arguments.
        :return: The name of the exchange, as it exists on the server.
        """

        async with self._checkout() as channel:
            return await channel.exchange_declare(
                name=name,
                type=type,
                passive=passive,
                durable=durable,
                auto_delete=auto_delete,
                internal=internal,
                arguments=arguments,
            )

    async def exchange_delete(self, name: str, *, if_unused: bool = False) -> None:
        """
        Deletes an exchange.
        :param name: The name of the exchange to delete.
        :param if_unused: If True, then the exchange will only be deleted if it has no queue
                          bindings.
        :return: Nothing.

        """

        async with self._checkout() as channel:
            await channel.exchange_delete(
                name=name,
                if_unused=if_unused,
            )

    async def exchange_unbind(
        self, destination: str, source: str, routing_key: str, **arguments: Any
    ) -> None:
        """
        Unbinds an exchange from another exchange. This is a
        `RabbitMQ extension <https://www.rabbitmq.com/e2e.html>`__ and may not be supported in other
        AMQP implementations.
        :param destination: The name of the destination exchange to unbind. A blank name means the
                            default exchange.
        :param source: The name of the source exchange to unbind. A blank name means the default
                       exchange.
        :param routing_key: The routing key for the exchange binding that is being unbinded.
        :param arguments: A dictionary of implementation-specific arguments.
        :return: Nothing.

        """

        async with self._checkout() as channel:
            await channel.exchange_unbind(
                destination=destination,
                source=source,
                routing_key=routing_key,
                arguments=arguments,
            )

    async def queue_bind(
        self, queue_name: str, exchange_name: str, routing_key: str, **arguments: Any
    ) -> None:
        """
        Binds a queue to an exchange.
        :param queue_name: The queue to bind.
        :param exchange_name: The exchange to bind to.
        :param routing_key: The routing key to use when binding.
        :param arguments: Any server-specific or exchange-specific extra arguments.
        :return: Nothing.

        """

        async with self._checkout() as channel:
            await channel.queue_bind(
                queue_name=queue_name,
                exchange_name=exchange_name,
                routing_key=routing_key,
                arguments=arguments,
            )

    async def queue_declare(
        self,
        name: str,
        *,
        passive: bool = False,
        durable: bool = False,
        exclusive: bool = False,
        auto_delete: bool = False,
        **arguments: Any
    ) -> QueueDeclareOkPayload:
        """
        Declares a queue.

        :param name: The name of the queue.
        :param passive: If True, the server will return a DeclareOk if the queue exists, and an
                        error if it doesn't. This can be used to inspect server state without
                        modification.
        :param durable: If True, the queue being created will persist past server restarts.
        :param exclusive: If True, this queue will only belong to this connection, and will be
                          automatically deleted when the connection closes. Best combined with an
                          automatically generated queue name.
        :param auto_delete: If True, this queue will be automatically deleted after all consumers
                            have finished. The queue will never be deleted before the first consumer
                            starts.
        :param arguments: Optional server implementation-specific arguments.
        :return: The :class:`.QueueDeclareOkPayload` the server returned.
        """

        async with self._checkout() as channel:
            return await channel.queue_declare(
                name=name,
                passive=passive,
                durable=durable,
                exclusive=exclusive,
                auto_delete=auto_delete,
                arguments=arguments,
            )

    async def queue_delete(
        self, queue_name: str, *, if_empty: bool = False, if_unused: bool = False
    ) -> int:
        """
        Deletes a queue.

        :param queue_name: The name of the queue to delete.
        :param if_empty: If True, the queue will only be deleted if it is empty.
        :param if_unused: If True, the queue will only be deleted if it is unused.
        :return: The number of messages deleted.
        """

        async with self._checkout() as channel:
            return await channel.queue_delete(
                queue_name=queue_name,
                if_empty=if_empty,
                if_unused=if_unused,
            )

    async def queue_purge(self, queue_name: str) -> int:
        """
        Purges all messages from a queue.

        :param queue_name: The name of the queue to be purged.
        :return: The number of messages deleted.
        """

        async with self._checkout() as channel:
            return await channel.queue_purge(
                queue_name=queue_name,
            )

    async def queue_unbind(
        self, queue_name: str, exchange_name: str, routing_key: str, **arguments: Any
    ) -> None:
        """
        Unbinds a queue from an exchange.

        :param queue_name: The name of the queue to unbind.
        :param exchange_name: The name of the exchange to unbind from.
        :param routing_key: The routing key to unbind using.
        :param arguments: Implementation-specific arguments to use.
        """

        async with self._checkout() as channel:
            return await channel.queue_unbind(
                queue_name=queue_name,
                exchange_name=exchange_name,
                routing_key=routing_key,
                arguments=arguments,
            )



from __future__ import annotations

from typing import Type, Union

from serena.enums import ReplyCode
from serena.payloads.method import (
    ChannelClosePayload,
    ConnectionClosePayload,
    MethodPayload,
    method_payload_name,
)

__all__ = (
    "AMQPError",
    "InvalidProtocolError",
    "AMQPStateError",
    "MessageReturnedError",
    "InvalidPayloadTypeError",
    "UnexpectedCloseError",
)


class AMQPError(Exception):
    """
    Base class exception for all AMQP-related exceptions.
    """

    __slots__ = ()


class InvalidProtocolError(AMQPError):
    """
    Thrown when the server speaks a protocol that we don't.
    """

    __slots__ = ()


class AMQPStateError(AMQPError):
    """
    Base exception for all AMQP state errors.
    """

    __slots__ = ()


class MessageReturnedError(AMQPError):
    """
    Thrown when a message is returned unexpectedly.
    """

    def __init__(self, reply_code: ReplyCode, reply_text: str, exchange: str, routing_key: str):
        #: The server-provided error code.
        self.reply_code = reply_code
        #: The server-provided error text.
        self.reply_text = reply_text

        #: The exchange the message was returned from.
        self.exchange = exchange

        #: The routing key the message was using.
        self.routing_key = routing_key

    def __str__(self):
        msg = (
            f"Message with routing key '{self.routing_key}' was returned from '{self.exchange}': "
            f"{self.reply_code.name}: {self.reply_text}"
        )

        return msg

    __repr__ = __str__


class InvalidPayloadTypeError(AMQPStateError):
    """
    Thrown when a payload's type is invalid.
    """

    __slots__ = ("expected", "actual")

    def __init__(self, expected: Type[MethodPayload], actual: MethodPayload):
        self.expected = expected
        self.actual = actual

        message = f"Expected {expected.__name__}, got {method_payload_name(actual)}"
        super().__init__(message)


class UnexpectedCloseError(AMQPError):
    """
    Thrown when the connection or a channeel closes unexpectedly.
    """

    __slots__ = ("reply_code", "reply_message", "class_id", "method_id")

    @classmethod
    def of(
        cls, payload: Union[ConnectionClosePayload, ChannelClosePayload]
    ) -> UnexpectedCloseError:
        """
        Creates a new :class:`.UnexpectedCloseError` from a close payload.
        """

        return UnexpectedCloseError(
            reply_code=payload.reply_code,
            reply_message=payload.reply_text,
            class_id=payload.class_id,
            method_id=payload.method_id,
        )

    def __init__(
        self,
        reply_code: ReplyCode,
        reply_message: str,
        class_id: int,
        method_id: int,
    ):
        #: The server-provided error code.
        self.reply_code = reply_code
        #: The server-provided error text.
        self.reply_message = reply_message

        #: The class ID of the method that caused this error.
        self.class_id = class_id
        #: The method ID of the method that caused this error.
        self.method_id = method_id

        message = f"{reply_code.name}: {reply_message}"
        if class_id > 0:
            message += f" (cause: {self.class_id}/{self.method_id})"

        super().__init__(message)



from __future__ import annotations

import logging
from contextlib import aclosing, asynccontextmanager
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    AsyncIterable,
    Awaitable,
    Callable,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

import anyio
from anyio import CancelScope, ClosedResourceError, EndOfStream, Event, Lock
from anyio.abc import TaskStatus
from anyio.lowlevel import checkpoint

from serena.enums import ExchangeType
from serena.exc import (
    AMQPStateError,
    InvalidPayloadTypeError,
    MessageReturnedError,
    UnexpectedCloseError,
)
from serena.frame import BodyFrame, Frame
from serena.message import AMQPEnvelope, AMQPMessage
from serena.mixin import ChannelLike
from serena.payloads.header import BasicHeader, ContentHeaderFrame, ContentHeaderPayload
from serena.payloads.method import (
    BasicAckPayload,
    BasicCancelPayload,
    BasicConsumeOkPayload,
    BasicConsumePayload,
    BasicDeliverPayload,
    BasicGetEmptyPayload,
    BasicGetOkPayload,
    BasicGetPayload,
    BasicNackPayload,
    BasicPublishPayload,
    BasicRejectPayload,
    BasicReturnPayload,
    ChannelClosePayload,
    ChannelOpenOkPayload,
    ExchangeBindOkPayload,
    ExchangeBindPayload,
    ExchangeDeclareOkPayload,
    ExchangeDeclarePayload,
    ExchangeDeleteOkPayload,
    ExchangeDeletePayload,
    ExchangeUnBindOkPayload,
    ExchangeUnBindPayload,
    MethodFrame,
    MethodPayload,
    QueueBindOkPayload,
    QueueBindPayload,
    QueueDeclareOkPayload,
    QueueDeclarePayload,
    QueueDeleteOkPayload,
    QueueDeletePayload,
    QueuePurgeOkPayload,
    QueuePurgePayload,
    QueueUnbindOkPayload,
    QueueUnbindPayload,
)

if TYPE_CHECKING:
    from serena.connection import AMQPConnection

_PAYLOAD = TypeVar("_PAYLOAD", bound=MethodPayload)

logger = logging.getLogger(__name__)


# noinspection PyProtectedMember
class Channel(ChannelLike):
    """
    A wrapper around an AMQP channel.
    """

    def __init__(self, channel_id: int, connection: AMQPConnection, stream_buffer_size: int):
        """
        :param channel_id: The ID of this channel.
        :param connection: The AMQP connection object to send data on.
        :param stream_buffer_size: The buffer size for the streams.
        """

        self._connection = connection
        self._channel_id = channel_id

        # Why is there both an `_open` and a `_closed`?
        # `_open` is True between ChannelOpenOk, and ChannelCloseOk, and is mostly just used for
        #  internal bookkeeping.
        # `_closed` is False when the ``async with`` block exits.

        self._open = False
        self._closed = False
        self._close_event = Event()

        # no buffer as these are for events that should return immediately
        self._send, self._receive = anyio.create_memory_object_stream(0)

        self._delivery_send, self._delivery_receive = anyio.create_memory_object_stream(
            max_buffer_size=stream_buffer_size
        )

        self._close_info: Optional[ChannelClosePayload] = None
        self._lock = Lock()

        self._is_consuming = False

        # internal state used by the connection
        # server requested a flow stop
        self._server_flow_stopped = False

        # used to count acks
        self._message_counter = 0

    def __str__(self):
        return f"<Channel id={self.id} buffered={self.current_buffer_size}>"

    __repr__ = __str__

    @property
    def id(self) -> int:
        """
        Returns the ID of this channel.
        """

        return self._channel_id

    @property
    def open(self) -> bool:
        """
        Returns if this channel is open or not.
        """

        return not self._closed

    @property
    def max_buffer_size(self) -> int:
        """
        Returns the maximum number of frames buffered in this channel. Used internally.
        """

        return int(self._delivery_send.statistics().max_buffer_size)

    @property
    def current_buffer_size(self) -> int:
        """
        Returns the current number of frames buffered in this channel. Used internally.
        """

        return self._delivery_send.statistics().current_buffer_used

    def _check_closed(self):
        """
        Checks if the channel is closed.
        """

        if self._closed:
            # todo: switch to our own exception?
            raise ClosedResourceError("This channel is closed")

    async def _close(self, payload: ChannelClosePayload):
        """
        Closes this channel.
        """

        self._close_info = payload
        await self._send.aclose()
        await self._delivery_send.aclose()
        self._closed = True

        # noinspection PyAsyncCall
        self._close_event.set()

        # aclose doesn't seem to checkpoint...
        await checkpoint()

    def _enqueue_regular(self, frame: MethodFrame):
        """
        Enqueues a regular method frame.
        """

        self._send.send_nowait(frame)

    async def _enqueue_delivery(self, frame: Frame):
        """
        Enqueues a delivery frame.
        """

        await self._delivery_send.send(frame)

    async def _receive_delivery_message(self):
        """
        Receives a single delivery message. This will reassemble a full message into its constituent
        frames.
        """

        method = None
        headers = None
        body = b""

        while True:
            # check for successful reassembly first
            if headers is not None and len(body) >= headers.payload.full_size:
                await checkpoint()
                # hehe payload.payload
                return AMQPMessage(
                    channel=self,
                    envelope=AMQPEnvelope.of(method.payload),
                    header=headers.payload.payload,
                    body=body,
                )

            try:
                next_frame = await self._delivery_receive.receive()
            except EndOfStream:
                if self._close_info is None:
                    raise AMQPStateError("Channel was closed improperly")

                raise UnexpectedCloseError.of(self._close_info) from None

            if method is None:
                if not isinstance(next_frame, MethodFrame):
                    raise AMQPStateError(f"Expected a method frame, got {next_frame} instead")

                if isinstance(next_frame.payload, (BasicGetOkPayload, BasicDeliverPayload)):
                    method = next_frame

                elif isinstance(next_frame.payload, BasicGetEmptyPayload):
                    return None

                else:
                    raise AMQPStateError(
                        f"Expected Basic.Deliver or Basic.Get-Ok, got {next_frame.payload} instead"
                    )

            elif headers is None:
                if not isinstance(next_frame, ContentHeaderFrame):
                    raise AMQPStateError(f"Expected a header frame, got {next_frame} instead")

                # explicit type hint as pycharm incorrectly infers based on the previous if check
                payload: ContentHeaderPayload = next_frame.payload  # type: ignore
                if payload.class_id != method.payload.klass:
                    raise AMQPStateError(
                        f"Class mismatch ({payload.class_id} != {method.payload.klass})"
                    )

                headers = next_frame

            else:
                if not isinstance(next_frame, BodyFrame):
                    raise AMQPStateError(f"Expected a body frame, got {next_frame} instead")

                body += next_frame.data

    async def _receive_frame(self) -> MethodFrame:
        """
        Receives a single frame from the channel.
        """

        self._check_closed()

        try:
            frame = await self._receive.receive()
        except EndOfStream:
            if self._close_info is None:
                raise AMQPStateError("Channel was closed improperly")

            raise UnexpectedCloseError.of(self._close_info) from None

        # notify connection code that we received an event so it can unblock the channel
        # noinspection PyAsyncCall
        return frame

    async def _wait_until_open(self):
        """
        Waits until the channel is open.
        """

        frame = await self._receive.receive()
        if not isinstance(frame, MethodFrame):
            raise AMQPStateError(f"Expected MethodFrame, got {frame}")

        if not isinstance(frame.payload, ChannelOpenOkPayload):
            raise InvalidPayloadTypeError(ChannelOpenOkPayload, frame.payload)

        self._open = True

    async def _send_single_frame(self, payload: MethodPayload):
        """
        Sends a single frame to the connection. This won't wait for a response.
        """

        async with self._lock:
            await self._connection._send_method_frame(self._channel_id, payload)

    async def _send_and_receive(
        self, callable: Callable[[], Awaitable[None]], type: Type[_PAYLOAD] = None
    ) -> MethodFrame[_PAYLOAD]:
        """
        Sends and receives a payload.

        :param callable: A callable to do the actual sending.
        :param type: The type of the expected payload to return.
        :return: A :class:`.MethodFrame` that was returned as a result.
        """

        returned_frame: MethodFrame[_PAYLOAD] = None  # type: ignore

        async def _closure(task_status: TaskStatus):
            task_status.started()

            nonlocal returned_frame
            returned_frame = await self._receive_frame()

        async with anyio.create_task_group() as nursery:
            await nursery.start(_closure)

            await callable()

        if type is not None and not isinstance(returned_frame.payload, type):
            raise InvalidPayloadTypeError(type, returned_frame.payload)

        return cast(MethodFrame[type], returned_frame)

    async def _send_and_receive_frame(
        self, payload: MethodPayload, type: Type[_PAYLOAD] = None
    ) -> MethodFrame[_PAYLOAD]:
        """
        Sends and receives a method payload.

        :param payload: The :class:`.MethodPayload` to send.
        :param type: The type of the expected payload to return.
        :return: A :class:`.MethodFrame` that was returned as a result.
        """

        async with self._lock:
            fn = partial(self._connection._send_method_frame, self._channel_id, payload)
            return await self._send_and_receive(fn, type=type)

    async def wait_until_closed(self):
        """
        Waits until the channel is closed.
        """

        await self._close_event.wait()

    ## METHODS ##
    async def exchange_declare(
        self,
        name: str,
        type: Union[ExchangeType, str],
        *,
        passive: bool = False,
        durable: bool = False,
        auto_delete: bool = False,
        internal: bool = False,
        **arguments: Any,
    ) -> str:
        """
        Declares a new exchange.

        :param name: The name of the exchange. Must not be empty.
        :param type: The type of the exchange to create.
        :param passive: If True, the server will return a DeclareOk if the exchange exists, and
                        an error if it doesn't. This can be used to inspect server state without
                        modification.
        :param durable: If True, then the declared exchange will survive a server restart.
        :param auto_delete: If True, then the declared exchange will be automatically deleted
                            when all queues have finished using it.
        :param internal: If True, then the exchange may not be used directly by publishers.
        :param arguments: A dictionary of implementation-specific arguments.
        :return: The name of the exchange, as it exists on the server.
        """

        if isinstance(type, ExchangeType):
            type = type.value

        payload = ExchangeDeclarePayload(
            reserved_1=0,
            name=name,
            exchange_type=type,
            passive=passive,
            durable=durable,
            auto_delete=auto_delete,
            internal=internal,
            no_wait=False,
            arguments=arguments,
        )

        await self._send_and_receive_frame(payload, ExchangeDeclareOkPayload)
        return payload.name

    async def exchange_delete(
        self,
        name: str,
        *,
        if_unused: bool = False,
    ) -> None:
        """
        Deletes an exchange.

        :param name: The name of the exchange to delete.
        :param if_unused: If True, then the exchange will only be deleted if it has no queue
                          bindings.
        :return: Nothing.
        """

        payload = ExchangeDeletePayload(
            reserved_1=0,
            name=name,
            if_unused=if_unused,
            no_wait=False,
        )

        await self._send_and_receive_frame(payload, ExchangeDeleteOkPayload)

    async def exchange_bind(
        self,
        destination: str,
        source: str,
        routing_key: str,
        **arguments: Any,
    ) -> None:
        """
        Binds an exchange to another exchange. This is a
        `RabbitMQ extension <https://www.rabbitmq.com/e2e.html>`__ and may not be supported in other
        AMQP implementations.

        :param destination: The name of the destination exchange to bind. A blank name means the
                            default exchange.
        :param source: The name of the source exchange to bind. A blank name means the default
                       exchange.
        :param routing_key: The routing key for the exchange binding.
        :param arguments: A dictionary of implementation-specific arguments.
        :return: Nothing.
        """

        payload = ExchangeBindPayload(
            reserved_1=0,
            destination_name=destination,
            source_name=source,
            routing_key=routing_key,
            no_wait=False,
            arguments=arguments,
        )

        await self._send_and_receive_frame(payload, ExchangeBindOkPayload)

    async def exchange_unbind(
        self,
        destination: str,
        source: str,
        routing_key: str,
        **arguments: Any,
    ) -> None:
        """
        Unbinds an exchange from another exchange. This is a
        `RabbitMQ extension <https://www.rabbitmq.com/e2e.html>`__ and may not be supported in other
        AMQP implementations.

        :param destination: The name of the destination exchange to unbind. A blank name means the
                            default exchange.
        :param source: The name of the source exchange to unbind. A blank name means the default
                       exchange.
        :param routing_key: The routing key for the exchange binding that is being unbinded.
        :param arguments: A dictionary of implementation-specific arguments.
        :return: Nothing.
        """

        payload = ExchangeUnBindPayload(
            reserved_1=0,
            destination_name=destination,
            source_name=source,
            routing_key=routing_key,
            no_wait=False,
            arguments=arguments,
        )

        await self._send_and_receive_frame(payload, ExchangeUnBindOkPayload)

    async def queue_declare(
        self,
        name: str = "",
        *,
        passive: bool = False,
        durable: bool = False,
        exclusive: bool = False,
        auto_delete: bool = False,
        **arguments: Any,
    ) -> QueueDeclareOkPayload:
        """
        Declares a queue.

        :param name: The name of the queue. If blank, a name will be automatically generated by
                     the server and returned.
        :param passive: If True, the server will return a DeclareOk if the queue exists, and an
                        error if it doesn't. This can be used to inspect server state without
                        modification.
        :param durable: If True, the queue being created will persist past server restarts.
        :param exclusive: If True, this queue will only belong to this connection, and will be
                          automatically deleted when the connection closes. Best combined with an
                          automatically generated queue name.
        :param auto_delete: If True, this queue will be automatically deleted after all consumers
                            have finished. The queue will never be deleted before the first consumer
                            starts.
        :param arguments: Optional server implementation-specific arguments.
        :return: The :class:`.QueueDeclareOkPayload` the server returned.
        """

        self._check_closed()

        payload = QueueDeclarePayload(
            reserved_1=0,
            name=name or "",
            passive=passive,
            durable=durable,
            exclusive=exclusive,
            auto_delete=auto_delete,
            no_wait=False,
            arguments=arguments or {},
        )

        result = await self._send_and_receive_frame(payload, QueueDeclareOkPayload)
        return result.payload

    async def queue_bind(
        self,
        queue_name: str,
        exchange_name: str,
        routing_key: str,
        **arguments: Any,
    ) -> None:
        """
        Binds a queue to an exchange.

        :param queue_name: The queue to bind.
        :param exchange_name: The exchange to bind to.
        :param routing_key: The routing key to use when binding.
        :param arguments: Any server-specific or exchange-specific extra arguments.
        :return: Nothing.
        """

        payload = QueueBindPayload(
            reserved_1=0,
            queue_name=queue_name,
            exchange_name=exchange_name,
            routing_key=routing_key,
            no_wait=False,
            arguments=arguments,
        )

        await self._send_and_receive_frame(payload, QueueBindOkPayload)

    async def queue_delete(
        self,
        queue_name: str,
        *,
        if_empty: bool = False,
        if_unused: bool = False,
    ) -> int:
        """
        Deletes a queue.

        :param queue_name: The name of the queue to delete.
        :param if_empty: If True, the queue will only be deleted if it is empty.
        :param if_unused: If True, the queue will only be deleted if it is unused.
        :return: The number of messages deleted.
        """

        payload = QueueDeletePayload(
            reserved_1=0,
            queue_name=queue_name,
            if_empty=if_empty,
            if_unused=if_unused,
            no_wait=False,
        )

        result = await self._send_and_receive_frame(payload, QueueDeleteOkPayload)
        return result.payload.message_count

    async def queue_purge(
        self,
        queue_name: str,
    ) -> int:
        """
        Purges all messages from a queue.

        :param queue_name: The name of the queue to be purged.
        :return: The number of messages deleted.
        """

        payload = QueuePurgePayload(
            reserved_1=0,
            name=queue_name,
            no_wait=False,
        )

        result = await self._send_and_receive_frame(payload, QueuePurgeOkPayload)
        return result.payload.message_count

    async def queue_unbind(
        self,
        queue_name: str,
        exchange_name: str,
        routing_key: str,
        **arguments: Any,
    ) -> None:
        """
        Unbinds a queue from an exchange.

        :param queue_name: The name of the queue to unbind.
        :param exchange_name: The name of the exchange to unbind from.
        :param routing_key: The routing key to unbind using.
        :param arguments: Implementation-specific arguments to use.
        """

        payload = QueueUnbindPayload(
            reserved_1=0,
            queue_name=queue_name,
            exchange_name=exchange_name,
            routing_key=routing_key,
            no_wait=False,
            arguments=arguments,
        )

        await self._send_and_receive_frame(payload, QueueUnbindOkPayload)

    def basic_consume(
        self,
        queue_name: str,
        consumer_tag: str = "",
        *,
        no_local: bool = False,
        no_ack: bool = False,
        exclusive: bool = False,
        auto_ack: bool = True,
        **arguments: Any,
    ) -> AsyncContextManager[AsyncIterable[AMQPMessage]]:
        """
        Starts a basic consume operation. This returns an async context manager over an asynchronous
        iterator that yields incoming :class:`.AMQPMessage` instances.

        The channel can still be used for other operations during this operation.

        :param queue_name: The name of the queue to consume from.
        :param consumer_tag: The tag for this consume.
        :param no_local: If True, messages will not be sent to this consumer if it is on the same
                         connection that published them.
        :param no_ack: If True, messages will not be expected to be acknowledged. This can cause
                       data loss.
        :param exclusive: If True, then only this consumer can access the queue. Will fail if there
                          is another consumer already active.
        :param arguments: Implementation-specific arguments.
        :param auto_ack: If True, then messages will be automatically positively acknowledged
                         in the generator loop. Has no effect if ``no_ack`` is True. This is a
                         Serena-exclusive feature, not a protocol feature.
        """

        if self._is_consuming:
            raise AMQPStateError("Cannot start two consumers on the same channel")

        async def _agen():
            while True:
                message = await self._receive_delivery_message()
                yield message

                if auto_ack:
                    await self.basic_ack(message.envelope.delivery_tag)

        @asynccontextmanager
        async def _do():
            payload = BasicConsumePayload(
                reserved_1=0,
                queue_name=queue_name,
                consumer_tag=consumer_tag,
                no_local=no_local,
                no_ack=no_ack,
                exclusive=exclusive,
                no_wait=False,
                arguments=arguments or {},
            )

            response = await self._send_and_receive_frame(payload, BasicConsumeOkPayload)

            try:
                self._is_consuming = True
                async with aclosing(_agen()) as agen:
                    yield agen
            finally:
                cancel_payload = BasicCancelPayload(
                    consumer_tag=response.payload.consumer_tag,
                    no_wait=False,
                )

                with CancelScope(shield=True):
                    if not self._closed:
                        await self._send_and_receive_frame(cancel_payload)

                    self._is_consuming = False

        return _do()

    async def basic_publish(
        self,
        exchange_name: str,
        routing_key: str,
        body: bytes,
        *,
        header: BasicHeader = None,
        mandatory: bool = True,
        immediate: bool = False,
    ):
        """
        Publishes a message to a specific exchange.

        :param exchange_name: The name of the exchange to publish to. This can be blank to mean the
                              default exchange.
        :param routing_key: The routing key to publish to.
        :param body: The body for this payload.
        :param header: The headers to use for this message. If unset, will use the default blank
                       headers.
        :param mandatory: Iff True, the server must return a ``Return`` message if the message
                          could not be routed to a queue.
        :param immediate: Iff True, the server must return a ``Return`` message if the message could
                          not be immediately consumed.
        :raise MessageReturnedError: If the message was returned to the publisher.

        .. warning::

            The immediate flag is *not* supported in RabbitMQ 3.x, and will cause the connection
            to close.
        """

        # we have to send three manual frames before we get an ACK

        self._check_closed()

        async with self._lock:
            method_payload = BasicPublishPayload(
                reserved_1=1,
                name=exchange_name,
                routing_key=routing_key,
                mandatory=mandatory,
                immediate=immediate,
            )

            # 1) method frame
            await self._connection._send_method_frame(self._channel_id, method_payload)

            # 2) header frame
            headers = header or BasicHeader()
            if body:
                # Body is not empty, we send_and_receive with the body and not the headers.
                await self._connection._send_header_frame(
                    self._channel_id,
                    method_klass=method_payload.klass,
                    body_length=len(body),
                    headers=headers,
                )

                # 3) Body
                fn = partial(self._connection._send_body_frames, self._channel_id, body)
                response = await self._send_and_receive(fn)

            else:
                # Body is empty, we send_and_receive with the headers.
                fn = partial(
                    self._connection._send_header_frame,
                    self._channel_id,
                    method_klass=method_payload.klass,
                    body_length=len(body),
                    headers=headers,
                )

                response = await self._send_and_receive(fn)

            self._message_counter += 1

            # 4) check for Ack or Return
            payload = response.payload

            if isinstance(payload, BasicAckPayload):
                if payload.delivery_tag != self._message_counter:
                    raise AMQPStateError(
                        f"Expected Ack for delivery tag {self._message_counter}, "
                        f"but got Ack for delivery tag {payload.delivery_tag}"
                    )
                logger.trace(f"C#{self.id}: Server ACKed published message")

                return

            elif isinstance(payload, BasicReturnPayload):
                raise MessageReturnedError(
                    exchange=payload.exchange,
                    routing_key=payload.routing_key,
                    reply_code=payload.reply_code,
                    reply_text=payload.reply_text,
                )

            elif isinstance(payload, BasicNackPayload):
                raise AMQPStateError("Server NACKed message to be published")

    async def basic_ack(self, delivery_tag: int, *, multiple: bool = False):
        """
        Acknowledges AMQP messages.

        :param delivery_tag: The delivery tag of the message to acknowledge.
        :param multiple: If True, then all messages up to and including the message specified will
                         be acknowledged, not just the message specified.
        """

        payload = BasicAckPayload(delivery_tag, multiple)
        await self._send_single_frame(payload)

    async def basic_reject(self, delivery_tag: int, *, requeue: bool = True):
        """
        Rejects an AMQP message.

        .. note::

            If you are using RabbitMQ, you might want to use :meth:`~.Channel.nack` instead.

        :param delivery_tag: The delivery tag of the message to acknowledge.
        :param requeue: If True, then the rejected message will be requeued if possible.
        """

        payload = BasicRejectPayload(delivery_tag, requeue)
        await self._send_single_frame(payload)

    async def basic_nack(
        self,
        delivery_tag: int,
        *,
        multiple: bool = False,
        requeue: bool = False,
    ):
        """
        Rejects an AMQP message. This is a RabbitMQ-specific extension.

        :param delivery_tag: The delivery tag of the message to acknowledge.
        :param multiple: If True, then all messages up to and including the message specified will
                         be acknowledged, not just the message specified.
        :param requeue: If True, then the rejected message will be requeued if possible.
        """

        payload = BasicNackPayload(delivery_tag, multiple, requeue)
        await self._send_single_frame(payload)

    async def basic_get(self, queue: str, *, no_ack: bool = False) -> Optional[AMQPMessage]:
        """
        Gets a single message from a queue.

        :param queue: The queue to get the message from.
        :param no_ack: Iff not True, then messages will need to be explicitly acknowledged on
                       consumption.
        :return: A :class:`.AMQPMessage` if one existed on the queue, otherwise None.
        """

        payload = BasicGetPayload(reserved_1=0, queue_name=queue, no_ack=no_ack)
        await self._send_single_frame(payload)
        return await self._receive_delivery_message()



from __future__ import annotations

import struct
from typing import Any, Dict, Optional, Tuple

import attr

from serena.frame import Frame, FrameType
from serena.payloads.encoding import (
    aq_type,
    decode_attrs_attribute,
    encode_attrs_attribute,
)
from serena.payloads.method import ClassID
from serena.utils.buffer import DecodingBuffer, EncodingBuffer

# whilst any class can define properties, Basic is the only one that does.
# if, in the future, other classes get properties then I will make this more generic.
# additionally, bit properties are packed differently but as there's no actual bit properties
# i skip the implementation.


@attr.s(frozen=True, slots=True)
class ContentHeaderFrame(Frame):
    """
    Frame for content headers.
    """

    type = FrameType.HEADER

    payload: ContentHeaderPayload = attr.ib()


@attr.s(frozen=True, slots=True)
class ContentHeaderPayload:
    #: The class ID of the content this header is for.
    class_id: ClassID = attr.ib()

    #: The full, complete size of the body payload.
    full_size: int = attr.ib()

    #: The payload flags.
    flags: int = attr.ib()

    #: The actual header data for this payload object.
    payload: BasicHeader = attr.ib()


@attr.s(frozen=True, slots=True)
class BasicHeader:
    """
    Defines the data for the Basic header.
    """

    #: The MIME content-type of this message.
    content_type: Optional[str] = attr.ib(default=None, kw_only=True)

    #: The MIME content-encoding of this message.
    content_encoding: Optional[str] = attr.ib(default=None, kw_only=True)

    #: The custom user headers for this message.
    headers: Dict[str, Any] = attr.ib(factory=dict, kw_only=True)

    #: The delivery mode for this message.
    delivery_mode: Optional[int] = attr.ib(default=None, kw_only=True, metadata=aq_type("octet"))

    #: The message priority, 0 through 9.
    priority: Optional[int] = attr.ib(default=None, kw_only=True, metadata=aq_type("octet"))

    #: The application correlation ID.
    correlation_id: Optional[str] = attr.ib(default=None, kw_only=True)

    #: The reply-to field.
    reply_to: Optional[str] = attr.ib(default=None, kw_only=True)

    #: The expiration of this message.
    expiration: Optional[str] = attr.ib(default=None, kw_only=True)

    #: The application-specific message ID of this message.
    message_id: Optional[str] = attr.ib(default=None, kw_only=True)

    #: The message timestamp.
    timestamp: Optional[int] = attr.ib(default=None, kw_only=True, metadata=aq_type("longlong"))

    #: The message type name.
    type_name: Optional[str] = attr.ib(default=None, kw_only=True)

    #: The creating user ID.
    user_id: Optional[str] = attr.ib(default=None, kw_only=True)

    #: The application ID.
    application_id: Optional[str] = attr.ib(default=None, kw_only=True)

    reserved_1: Optional[str] = attr.ib(default=None, kw_only=True)


def deserialise_basic_header(payload: bytes) -> ContentHeaderPayload:
    """
    Deserialises the Basic content header into a payload object.
    """

    klass, body_size, flags, rest = (
        ClassID(int.from_bytes(payload[0:2], byteorder="big")),
        int.from_bytes(payload[4:12], byteorder="big"),
        int.from_bytes(payload[12:14], byteorder="big"),
        payload[14:],
    )

    attr.resolve_types(BasicHeader)
    fields: Tuple[attr.Attribute] = attr.fields(BasicHeader)
    buffer = DecodingBuffer(rest)

    params = {}

    for idx, field in enumerate(fields):
        bit_idx = len(fields) - (idx - 1)
        is_provided = (flags & (1 << bit_idx)) != 0
        if is_provided:
            data = decode_attrs_attribute(buffer, field)
            params[field.name] = data

    header = BasicHeader(**params)
    return ContentHeaderPayload(class_id=klass, full_size=body_size, flags=flags, payload=header)


def serialise_basic_header(klass_id: ClassID, body_size: int, body: BasicHeader):
    """
    Serialises a Basic content header into a byte array.

    :param klass_id: The class ID of the method the header is for.
    :param body_size: The full size of the body.
    :param body: The header body.
    """

    attr.resolve_types(BasicHeader)
    fields: Tuple[attr.Attribute] = attr.fields(BasicHeader)
    buffer = EncodingBuffer()

    flags = 0

    for idx, field in enumerate(fields):
        field_value = getattr(body, field.name)
        if field_value:  # truthy check for the empty dict
            bit_idx = len(fields) - (idx - 1)
            flags |= 1 << bit_idx
            encode_attrs_attribute(buffer, field, field_value)

    header = struct.pack(">HHQH", klass_id.value, 0, body_size, flags)
    return header + buffer.get_data()



from __future__ import annotations

import logging
import struct
from math import ceil
from typing import List, Union

from serena.frame import BodyFrame, Frame, HeartbeatFrame
from serena.payloads.header import (
    BasicHeader,
    ContentHeaderFrame,
    deserialise_basic_header,
    serialise_basic_header,
)
from serena.payloads.method import (
    ClassID,
    MethodFrame,
    MethodPayload,
    deserialise_payload,
    method_payload_name,
    serialise_payload,
)


class _NEED_DATA:
    pass


#: A special singleton object used if the parser needs more data.
NEED_DATA = _NEED_DATA()

logger = logging.getLogger(__name__)

# Dear RabbitMQ authors:
# Fuck you. I fucking HATE you. Implement the FUCKING SPEC PROPERLY.
# I love spending a day of my life implementing the AMQP 0-9-1 spec only to find that RabbitMQ
# fucking does something else. Short strings in tables? Thrown away. Heartbeat frames? Type 8
# instead of type 4. What the fuck??? Why?????????

METHOD_FRAME = 1
HEADER_FRAME = 2
BODY_FRAME = 3
HEARTBEAT_FRAME = 8


class FrameParser(object):
    """
    A buffer that parses AMQP frames. This is not an AMQP state machine, and is only suitable for
    the low-level wire processing.
    """

    def __init__(self):
        self._buffer: bytearray = bytearray()
        self._processing_partial_packet = False

        # saved state
        self._saved_type = -1
        self._saved_channel = -1
        self._saved_size = -1
        self._bytes_remaining = -1
        self._last_packet_buffer = bytearray()

    @staticmethod
    def _make_frame(type: int, channel: int, payload: bytes) -> Frame:
        """
        Creates a new frame object.
        """

        if type == METHOD_FRAME:
            decoded_payload = deserialise_payload(payload)
            frame = MethodFrame(channel_id=channel, payload=decoded_payload)
            logger.trace(
                f"S#{channel}->C FRAME (METHOD): {method_payload_name(decoded_payload)} "
                f"({len(payload)}B)"
            )
            return frame

        elif type == HEADER_FRAME:
            decoded_payload = deserialise_basic_header(payload)
            logger.trace(
                f"S#{channel}->C FRAME (HEADER): {decoded_payload.class_id.name} ("
                f"{len(payload)} bytes)"
            )
            frame = ContentHeaderFrame(channel_id=channel, payload=decoded_payload)
            return frame

        elif type == BODY_FRAME:
            logger.trace(f"S#{channel}->C FRAME (BODY): {len(payload)} bytes")
            frame = BodyFrame(channel_id=channel, data=payload)
            return frame

        elif type == HEARTBEAT_FRAME:
            assert channel == 0, "heartbeats cannot be on any channel other than zero"
            logger.trace("S->C FRAME (HEARTBEAT)")
            return HeartbeatFrame(channel_id=0)

        else:
            raise ValueError(f"Invalid frame type: {type}")

    @staticmethod
    def _pack_frame(type_: int, channel: int, payload: bytes) -> bytes:
        """
        Packs a single frame with a header.

        :param type_: The type of the frame.
        :param channel: The channel the frame is on.
        :param payload: The opaque payload body.
        :return: A fully encoded frame.
        """

        size = len(payload)
        header = struct.pack(">BHI", type_, channel, size)
        result = header + payload + b"\xCE"
        return result

    @staticmethod
    def write_method_frame(channel: int, payload: MethodPayload) -> bytes:
        """
        Writes a method payload into a bytestring.

        :param channel: The channel ID this payload is being sent on.
        :param payload: The payload object to send.
        :return: The :class:`bytes` to send over the wire.
        """

        body = serialise_payload(payload)
        result = FrameParser._pack_frame(METHOD_FRAME, channel, body)
        logger.trace(
            f"C#{channel}->S FRAME (METHOD): {method_payload_name(payload)} ({len(body)} bytes)"
        )
        return result

    @staticmethod
    def write_header_frame(
        channel: int, method_klass: ClassID, body_length: int, headers: BasicHeader
    ) -> bytes:
        """
        Writes a header payload into a bytearray.
        """

        frame_body = serialise_basic_header(method_klass, body_length, headers)
        result = FrameParser._pack_frame(HEADER_FRAME, channel, frame_body)
        logger.trace(f"C{channel}->S FRAME (HEADER): {method_klass.name} ({len(frame_body)} bytes)")
        return result

    @staticmethod
    def write_body_frames(channel: int, body: bytes, max_frame_size: int) -> List[bytes]:
        """
        Writes a set of body frames into a set of byte strings.
        """

        if len(body) <= 0:
            return []

        frames_needed = ceil(len(body) / max_frame_size)
        frames = []

        for i in range(0, frames_needed):
            frame_body = body[max_frame_size * i : max_frame_size * (i + 1)]
            frame = FrameParser._pack_frame(BODY_FRAME, channel, frame_body)
            logger.trace(
                f"C{channel}->S FRAME (BODY): {i + 1} / {frames_needed} ({len(frame_body)} bytes)"
            )
            frames.append(frame)

        return frames

    def receive_data(self, data: bytes):
        """
        Receives incoming data from the AMQP server.
        """

        self._buffer += data

    def next_frame(self) -> Union[Frame, _NEED_DATA]:
        """
        Retrieves the next frame from the buffer.
        """

        if len(self._buffer) <= 0:
            return NEED_DATA

        if self._processing_partial_packet:
            body, self._buffer = (
                self._buffer[: self._bytes_remaining],
                self._buffer[self._bytes_remaining :],
            )
            self._last_packet_buffer += body
            self._bytes_remaining -= len(body)

            if self._bytes_remaining > 0:
                logger.trace(
                    f"Still missing {self._bytes_remaining} bytes from packet, skipping cycle"
                )
                return NEED_DATA

            # truncate frame-end octet
            assert body[-1] == 0xCE, "invalid frame-end octet"
            body = body[:-1]

            # packet finished, construct frame from saved values
            self._processing_partial_packet = False
            frame = self._make_frame(
                self._saved_type, self._saved_channel, self._last_packet_buffer
            )
            self._last_packet_buffer = bytearray

            return frame
        else:
            # header is 7 octets, without it we don't care.
            if len(self._buffer) < 7:  # pragma: no cover
                logger.trace("Missing packet header, skipping")
                return NEED_DATA

            # pop off bits
            type_ = self._buffer[0]
            channel = (self._buffer[1] << 8) | self._buffer[2]
            size = (
                self._buffer[3] << 24
                | self._buffer[4] << 16
                | self._buffer[5] << 8
                | self._buffer[6]
            ) + 1  # + 1 is for the frame end byte (0xCE)

            logger.trace(f"Received packet ({type_=} | {channel=} | {size=})")

            body, self._buffer = self._buffer[7 : size + 7], self._buffer[size + 7 :]

            if len(body) < size:
                # missing part of the packet, save and return later
                self._processing_partial_packet = True
                self._saved_size = size
                self._saved_channel = channel
                self._saved_type = type_
                self._last_packet_buffer = body
                self._bytes_remaining = size - len(body)
                logger.trace(f"Missing {self._bytes_remaining} bytes from packet, skipping cycle")
                return NEED_DATA
            else:
                assert body[-1] == 0xCE, "invalid frame-end octet"
                body = body[:-1]

                return self._make_frame(type_, channel, body)



from enum import IntEnum

try:
    from enum import StrEnum
except ImportError:
    from backports.strenum import StrEnum


__all__ = (
    "ReplyCode",
    "ExchangeType",
    "ClassID",
)


class ClassID(IntEnum):
    """
    Enumeration of method class IDs.
    """

    CONNECTION = 10
    CHANNEL = 20
    EXCHANGE = 40
    QUEUE = 50
    BASIC = 60
    CONFIRM = 85
    TX = 90


class ReplyCode(IntEnum):
    """
    Enumeration of possible AMQP reply codes.
    """

    #: Indicates that the method completed successfully. This reply code is reserved for future use
    #: - the current protocol design does not use positive confirmation and reply codes are sent
    #: only in case of an error.
    success = 200

    #: The client attempted to transfer content larger than the server could accept at the present
    #: time. The client may retry at a later time.
    content_too_large = 311

    #: Message cannot be delivered to any queue.
    no_route = 312

    #: When the exchange cannot deliver to a consumer when the immediate flag is set. As a result
    #: of pending data on the queue or the absence of any consumers of the queue.
    no_consumers = 313

    #: An operator intervened to close the connection for some reason. The client may retry at
    #: some later date.
    connection_forced = 320

    #: The client tried to work with an unknown virtual host.
    invalid_path = 402

    #: The client attempted to work with a server entity to which it has no access due to security
    #: settings.
    access_refused = 403

    #: The client attempted to work with a server entity that does not exist.
    not_found = 404

    #: The client attempted to work with a server entity to which it has no access because
    #: another client is working with it.
    resource_locked = 405

    #: The client requested a method that was not allowed because some precondition failed.
    precondition_failed = 406

    #: The sender sent a malformed frame that the recipient could not decode. This strongly implies
    #: a programming error in the sending peer.
    frame_error = 501

    #: The sender sent a frame that contained illegal values for one or more fields. This strongly
    #: implies a programming error in the sending peer.
    syntax_error = 502

    #: The client sent an invalid sequence of frames, attempting to perform an operation that was
    #: considered invalid by the server. This usually implies a programming error in the client.
    command_invalid = 503

    #: The client attempted to work with a channel that had not been correctly opened. This most
    #: likely indicates a fault in the client layer.
    channel_error = 504

    #: The peer sent a frame that was not expected, usually in the context of a content header
    #: and body. This strongly indicates a fault in the peer's content processing.
    unexpected_frame = 505

    #: The server could not complete the method because it lacked sufficient resources. This may
    #: be due to the client creating too many of some type of entity.
    resource_error = 506

    #: The client tried to work with some entity in a manner that is prohibited by the server, due
    #: to security settings or by some other criteria.
    not_allowed = 530

    #: The client tried to use functionality that is not implemented in the server.
    not_implemented = 540

    #: The server could not complete the method because of an internal error. The server may
    #: require intervention by an operator in order to resume normal operations.
    internal_error = 541


class ExchangeType(StrEnum):
    """
    Enumeration of possible exchange types.
    """

    DIRECT = "direct"
    FANOUT = "fanout"
    TOPIC = "topic"
    HEADERS = "headers"
